# -*- coding: utf-8 -*-
# Copyright (c) 2019 NewStore Inc. <engineering@newstore.com>
#
# this file is part of the project "Mercator - Data Mapper for Protobuf" released under the "MIT" open-source license

version = '0.1.8'
